@@ -23,6 +23,8 @@ from nameko.standalone.events import event_dispatcher
 from nameko.constants import AMQP_URI_CONFIG_KEY, HEARTBEAT_CONFIG_KEY
 from six.moves import queue as queue_six
 from six.moves import xrange as xrange_six
+import atexit
+import math
 
 _logger = logging.getLogger(__name__)
 
@@ -108,13 +110,13 @@ class ClusterRpcProxyPool(object):
                     if self._rpc._worker_ctx.data is not None:
                         if self._pool.context_data is None:
                             # clear all key since there is no.pool context_data
-                            for key in self._rpc._worker_ctx.data.keys():
+                            for key in list(self._rpc._worker_ctx.data.keys()):
                                 del self._rpc._worker_ctx.data[key]
                         elif len(self._rpc._worker_ctx.data) != len(self._pool.context_data) \
                                 or self._rpc._worker_ctx.data != self._pool.context_data:
                             # ensure that worker_ctx.data is revert back to original
                             # pool.context_data when exit of block
-                            for key in self._rpc._worker_ctx.data.keys():
+                            for key in list(self._rpc._worker_ctx.data.keys()):
                                 if key not in self._pool.context_data:
                                     del self._rpc._worker_ctx.data[key]
                                 else:
@@ -151,6 +153,7 @@ class ClusterRpcProxyPool(object):
         self.state = 'STARTED'
         if self.heartbeat:
             self._heartbeat_check_thread = Thread(target=self.heartbeat_check)
+            self._heartbeat_check_thread.daemon = True
             self._heartbeat_check_thread.start()
             _logger.debug("Heart beat check thread started")
 
@@ -197,6 +200,7 @@ class ClusterRpcProxyPool(object):
     def stop(self):
         """ Stop queue and remove all connections from pool.
         """
+        self.state = 'STOPPED'
         if self.queue:
             while True:
                 try:
@@ -206,42 +210,81 @@ class ClusterRpcProxyPool(object):
                     break
             self.queue.queue.clear()
             self.queue = None
-        self.state = 'STOPPED'
-        if self._heartbeat_check_thread:
-            self._heartbeat_check_thread.join()
-            _logger.debug("Heart beat check thread stopped")
+        # if self._heartbeat_check_thread:
+        #     self._heartbeat_check_thread.join()
+        #     _logger.debug("Heart beat check thread stopped")
 
     def heartbeat_check(self):
-        RATE = 2
+        RATE = 2 + math.log(self.heartbeat, 30) if self.heartbeat > 30 else 2.
+        MIN_SLEEP = 3  # better sleep between 3 seconds, if this loop is running too frequent it may affect performance
+        loop_count = 0
+        REPLIES_CLEAN_UP_CYCLE = 10  # how many loop cycle to perform replies clean up
+        replies_timestamp = {}  # hash of correlation_id of replies and its timestamp when first detected
+        sleep_duration = max(self.heartbeat / abs(RATE), MIN_SLEEP)
         while self.heartbeat and self.state == 'STARTED':
-            time.sleep(self.heartbeat / abs(RATE))
+            for i in range(10):
+                # breakdown sleep in smaller step
+                time.sleep(sleep_duration/10.0)
+                if self.state == 'STOPPED':
+                    return
             count_ok = 0
-            cleared = list()
-            for _ in xrange_six(self.pool_size):
-                ctx = None
-                try:
-                    ctx = self.queue.get_nowait()
-                except queue_six.Empty:
-                    break
-                else:
-                    if ctx._rpc and id(ctx) not in cleared:
-                        try:
+            cleared = set()
+            try:
+                for _ in xrange_six(self.pool_size):
+                    if self.state == 'STOPPED':
+                        return
+                    ctx = None
+                    try:
+                        ctx = self.queue.get_nowait()
+                    except (queue_six.Empty, AttributeError):
+                        break
+                    else:
+                        if ctx and ctx._rpc and id(ctx) not in cleared:
                             try:
-                                ctx._rpc._reply_listener.queue_consumer.connection.drain_events(timeout=0.1)
-                            except socket.timeout:
-                                pass
-                            ctx._rpc._reply_listener.queue_consumer.connection.heartbeat_check()  # rate=RATE
-                        except (ConnectionError, socket.error) as exc:
-                            _logger.info("Heart beat failed. System will auto recover broken connection: %s", str(exc))
+                                try:
+                                    ctx._rpc._reply_listener.queue_consumer.connection.drain_events(timeout=0.1)
+                                except socket.timeout:
+                                    pass
+                                ctx._rpc._reply_listener.queue_consumer.connection.heartbeat_check()
+                            except (ConnectionError, socket.error, IOError) as exc:
+                                _logger.info("Heart beat failed. System will discard broken connection and replenish "
+                                             "pool with a new connection, %s: %s",
+                                             type(exc).__name__, exc.args[0])
+                                ctx.__del__()
+                                ctx = ClusterRpcProxyPool.RpcContext(self, self.config)
+                            else:
+                                count_ok += 1
+                                # this allow only one RPCProxy connection to be cleanup at a time
+                                d = loop_count - count_ok
+                                if d > 0 and d % REPLIES_CLEAN_UP_CYCLE == 0:
+                                    count_clean = 0
+                                    now = time.time()
+                                    # perform cleanup on this RpcProxy connection replies
+                                    for msg_correlation_id in list(ctx._rpc._reply_listener.queue_consumer.replies.keys()):
+                                        if self.state == 'STOPPED':
+                                            return
+                                        timestamp = replies_timestamp.get(msg_correlation_id)
+                                        if timestamp is None:
+                                            replies_timestamp[msg_correlation_id] = now
+                                        else:
+                                            # clean up the reply if its has stay in replies
+                                            if now - timestamp > self.timeout:
+                                                del ctx._rpc._reply_listener.queue_consumer.replies[msg_correlation_id]
+                                                del replies_timestamp[msg_correlation_id]
+                                                count_clean += 1
+                                    _logger.debug("Perform cleanup remove %d message", count_clean)
+                    finally:
+                        if ctx is not None and self.queue is not None:
+                            self.queue.put_nowait(ctx)
+                            cleared.add(id(ctx))
+                        elif ctx is not None:
+                            #  unable to put it back, probaly due to system exit so better just delete the connection
                             ctx.__del__()
-                            ctx = ClusterRpcProxyPool.RpcContext(self, self.config)
-                        else:
-                            count_ok += 1
-                finally:
-                    if ctx is not None:
-                        self.queue.put_nowait(ctx)
-                        cleared.append(id(ctx))
+            except Exception as exc:
+                _logger.error("%s: %s", type(exc).__name__, exc.args[0])
+                # just log the error out without raise to keep the heartbeat thread going
             _logger.debug("Heart beat %d OK", count_ok)
+            loop_count += 1
 
     def __del__(self):
         if self.state != 'STOPPED':
@@ -353,6 +396,7 @@ def get_pool(pool_name=None):
     return _pool
 
 
+@atexit.register
 def destroy_pool():
     global nameko_global_pools
     if isinstance(nameko_global_pools, dict):
@@ -361,6 +405,7 @@ def destroy_pool():
     elif nameko_global_pools is not None:
         nameko_global_pools.stop()
     nameko_global_pools = None
+    _logger.info("nameko_global_pools are destroyed")
 
 
 # setup a singleton event dispatcher for current worker
@@ -375,13 +420,13 @@ def get_event_dispatcher():
         if not NAMEKO_CONFIG:
             raise ImproperlyConfigured('NAMEKO_CONFIG must be specified')
         # Lazy instantiation, acquire lock first to prevent dupication init
-        create_event_dispatcher_lock.acquire()
-        _logger.debug("init nameko_event_dispatcher")
-        nameko_event_dispatcher = event_dispatcher(
-            NAMEKO_CONFIG['default'] if 'default' in NAMEKO_CONFIG else NAMEKO_CONFIG
-        )
-        # Finish instantiation, release lock
-        create_event_dispatcher_lock.release()
+        with create_event_dispatcher_lock:
+            if not nameko_event_dispatcher:  # double check inside lock is importance
+                # init nameko_event_dispatcher
+                nameko_event_dispatcher = event_dispatcher(
+                    NAMEKO_CONFIG['default'] if 'default' in NAMEKO_CONFIG else NAMEKO_CONFIG
+                )
+            # Finish instantiation, lock will be released automaticaly when exit this block
     return nameko_event_dispatcher
 
 
